package at.ac.uibk.gitsearch.service;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.zip.ZipInputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.FileCopyUtils;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
public class SearchService {

	private final Logger log = LoggerFactory.getLogger(SearchService.class);

	@Autowired
	protected MetaDataRepository metaDataRepository;
	@Autowired
	protected TokenProvider tokenProvider;
	@Autowired
	protected GitLabRepository gitLabRepository;
	@Autowired
	protected PluginManagementService pluginManagementService;
	@Autowired
	protected ShoppingBasketService shoppingBasketService;

	/** just for test implementation **/
	public static final int NUM_TESTRESULTS = 23;

	public SearchService() {
		// Just for Spring
	}

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix) throws IOException {
		return metaDataRepository.getKeywordsAutoComplete(keyWordPrefix);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix) throws IOException {
		return metaDataRepository.getCreatorAutoComplete(creatorPrefix);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix) throws IOException {
		return metaDataRepository.getContributorAutoComplete(contributorPrefix);
	}

	/**
	 * returns all contributor and author autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix) throws IOException {
		return metaDataRepository.getContributorCreatorAutoComplete(contributorPrefix);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param programmingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String programmingLanguagePrefix)
			throws IOException {
		return metaDataRepository.getProgrammingLanguageAutoComplete(programmingLanguagePrefix);
	}

	/**
	 * returns the result page
	 *
	 * @param searchInput the query definition
	 * @param first       the index of the first record to be returned
	 * @param length      the number of records
	 */
	public SearchResultsDTO searchResultPage(SearchInputDTO searchInput, long first, int length) throws IOException {
		log.debug("Searchrequest for {} ", searchInput);

		final SearchResultsDTO pageDetails = metaDataRepository.pageDetails(searchInput, length,
				tokenProvider.getCurrentPrincipal());

		pageDetails.getSearchResult().stream()
				.forEach(hit -> pluginManagementService.getRegisteredPluginConfigs().stream()
						.forEach(config -> config.getActions().values().stream()
								.filter(action -> action.isApplicable(hit))
								.forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))));
		pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

		return pageDetails;
		// return readTestResults(searchInput.getFulltextQuery(), first, length);
	}

	private void fixImageURL(SearchResultDTO metaData) {

		String image = metaData.getMetadata().getImage();

		try {
			if (image != null) {
				URI url = new URI(image);
				if (url.isAbsolute())
					return;
				String httpUrlToRepo = metaData.getProject().getUrl();
				String baseRepoURL = httpUrlToRepo + "/-/raw/master/";
				final URI resolvedImageUrl = new URI(baseRepoURL).resolve(url);
				metaData.getMetadata().setImage(resolvedImageUrl.toASCIIString());
				return;
			}

			final String baseUrl = ServletUriComponentsBuilder.fromCurrentContextPath().build().toUriString();
			final boolean isJava = metaData.getMetadata().getProgrammingLanguage() != null && Arrays
					.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("JAVA"))
					&& metaData.hashCode() % 2 == 0; // nur um ein wenig Abwechslung zu bekommen :-)
			final boolean isLatexFormat = metaData.getMetadata().getFormat() != null
					&& Arrays.stream(metaData.getMetadata().getFormat()).anyMatch(s -> s.startsWith("latex"));
			final boolean isPython = metaData.getMetadata().getProgrammingLanguage() != null && Arrays
					.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("python"));
			if (isJava) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/java.png");
				return;
			}
			if (isLatexFormat) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/latex.png");
				return;
			}
			if (isPython) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/python.png");
				return;
			}
			metaData.getMetadata().setImage(baseUrl + "/content/img/gitLab.png");
			return;

		} catch (URISyntaxException e) {
			log.warn("Cannot parse image url {} for {}", metaData.getMetadata().getImage(),
					metaData.getProject().getUrl());
		}
	}

	/**
	 * temporary static test data.
	 *
	 * @param searchString just a string to add to testdata description.
	 * @return
	 */
	@SuppressWarnings("unused")
	private SearchResultsDTO readTestResults(String searchString, long first, int length) {
		ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
		mapper.findAndRegisterModules();

		List<SearchResultDTO> loadedMetaData = new ArrayList<>();

		for (String testFile : new String[] { "metaData1.yaml", "metaData2.yaml" }) {
			log.debug("reading test data from {} ", testFile);

			final File metaDataFile = new File("src/main/java/at/ac/uibk/gitsearch/service/testData/" + testFile);
			try {
				byte[] fileContent = FileCopyUtils.copyToByteArray(metaDataFile);
				SearchResultDTO searchResultDTO = mapper.readValue(fileContent, SearchResultDTO.class);
				loadedMetaData.add(searchResultDTO);
			} catch (IOException e) {
				log.error("Cannot read test search input from " + testFile, e);
			}
		}

		List<SearchResultDTO> result = new ArrayList<>();
		if (length > 0) {
			int metadataPos = 0;
			long recordNr = first;
			// This loop is only used for testing to get more results.
			for (int i = 0; i < 12; ++i) {
				metadataPos = 0; // also for testing, remove later
				while (metadataPos < length && metadataPos < loadedMetaData.size()) {
					final SearchResultDTO loadedMetaDataRecord = new SearchResultDTO(loadedMetaData.get(metadataPos++));
					if (loadedMetaDataRecord.toString().contains(searchString)) {
						loadedMetaDataRecord.getMetadata()
								.setTitle("#" + (recordNr + 1) + " " + loadedMetaDataRecord.getMetadata().getTitle());
						result.add(loadedMetaDataRecord);
						recordNr++;
						log.debug("Added MetaData Record {}", loadedMetaData);
					}
					log.debug("{} does not contain {}", loadedMetaData, searchString);
				}
			}

		}
		return new SearchResultsDTO(result, NUM_TESTRESULTS, first);

	}

	public File exportExercise(long exerciseId) throws IOException {
		try{
		final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
		InputStream inputFile = shoppingBasketService.rePackageGitLabProjectZip(new ZipInputStream(gitLabApi.getRepositoryApi().getRepositoryArchive((int)(long)exerciseId, "HEAD", "zip")), "from project " + String.valueOf(exerciseId));
		File file = new File("exercise" + String.valueOf(exerciseId) + ".zip");

		return copyInputStreamToFile(inputFile, file);}
		catch(GitLabApiException exception){
			log.error(exception.getMessage());
			return null;
		}
	}

	private File copyInputStreamToFile(InputStream inputStream, File file) throws IOException {

		// append = false
		try (FileOutputStream outputStream = new FileOutputStream(file, false)) {
			int read;
			byte[] bytes = new byte[8192];
			while ((read = inputStream.read(bytes)) != -1) {
				outputStream.write(bytes, 0, read);
			}
			return file;
		}
	}

}
