package at.ac.uibk.ac.artemis;

import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.junit.Test;

import at.ac.uibk.ac.artemis.testutils.InterfaceProxy;

/**
 * @author Michael Breu
 * @version 1.0 (06.05.20)
 */
public class LeakyBehaviorTest extends BehaviorTest {

	private static final String POTENTIALLY_FAULTY_PACKET_INTERFACENAME = "at.ac.uibk.ac.artemis.LeakyBucketPatternInterface$PotentiallyFaultyPacket";

	/**
	 * wraps a receiver.
	 */
	public class ReceiverWrapper {

		Object receiver;

		public ReceiverWrapper() throws IllegalAccessException, IllegalArgumentException, InvocationTargetException,
				NoSuchMethodException, SecurityException, InstantiationException, ClassNotFoundException {

			receiver = newInstance("at.ac.uibk.ac.artemis.Receiver");
		}

		/**
		 * initializes the receiver, to throw an error, if the given error Rate is
		 * exceeded.
		 * 
		 * @param errorRatePer_msec the rate of faulty packets per second.
		 * @throws IOException
		 */
		public void initReceiver(float errorRatePerSec) throws Exception {
			try {
				invokeMethod(receiver, getMethod(receiver, "initReceiver", float.class), errorRatePerSec);
			} catch (Throwable e) {
				throw new InvocationTargetException(e);
			}
		}

		Class<?> interfaceClass = this.getClass().getClassLoader().loadClass(POTENTIALLY_FAULTY_PACKET_INTERFACENAME);

		/**
		 * interface to receive a packet.
		 *
		 * @param potentiallyFaultyPacket the packet to receive
		 */
		public void receivePacket(Object potentiallyFaultyPacket) throws Exception {
			try {
				invokeMethod(receiver, getMethod(receiver, "receivePacket", interfaceClass), potentiallyFaultyPacket);
			} catch (Throwable e) {
				if (e instanceof IOException) {
					IOException ioException = (IOException) e;
					throw ioException;
				} else {
					throw new InvocationTargetException(e);
				}
			}
		}

	}

	/**
	 * just simulates a set of streams
	 * 
	 * @author Michael Breu
	 *
	 */
	public class TestStream {

		private final class PotentialFaultyPacketProxyHandler implements InvocationHandler {

			private final PacketDefinition packetDefinition;

			public PotentialFaultyPacketProxyHandler(PacketDefinition packetDefinition) {
				super();
				this.packetDefinition = packetDefinition;
			}

			@Override
			public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
				String name = method.getName();

				if (name.equals("toString")) {

					return "toString for " + packetDefinition + " was called.";
				} else if (name.equals("hashCode")) {

					return packetDefinition.hashCode();
				} else if (name.equals("payLoad")) {
					return new byte[] {};
				} else if (name.equals("isFaulty")) {
					return packetDefinition.faulty;
				}

				throw new RuntimeException("no method found");
			}
		}

		private class PacketDefinition {
			long pauseInterval;
			boolean faulty;

			private PacketDefinition(long pauseInterval, boolean faulty) {
				this.pauseInterval = pauseInterval;
				this.faulty = faulty;
			}
		}

		private List<PacketDefinition> packetStreamDefinition = new ArrayList<PacketDefinition>();

		private TestStream(String definition) {
			StringTokenizer st = new StringTokenizer(definition);

			while (st.hasMoreTokens()) {
				String interval = st.nextToken();
				if (!st.hasMoreTokens()) {
					throw new IllegalArgumentException("Keine gerade Anzahl von Einträgen);");
				}
				String type = st.nextToken();

				long pauseInterval = Long.parseLong(interval);
				boolean faulty = type.startsWith("F") || type.startsWith("f");

				packetStreamDefinition.add(new PacketDefinition(pauseInterval, faulty));
			}
		}

		private void simulateStream(ReceiverWrapper leakyBucket) throws Exception {
			long lastPacketSent = System.currentTimeMillis();

			for (PacketDefinition packetDefinition : packetStreamDefinition) {
				// just to adjust some time inaccuracies.
				long wait = lastPacketSent + packetDefinition.pauseInterval - System.currentTimeMillis();
				if (wait > 0) {
					// this is a very lazy implementation of wait. For testing it should be ok.
					Thread.currentThread().sleep(wait);
				}
				leakyBucket.receivePacket(InterfaceProxy.getInstance(POTENTIALLY_FAULTY_PACKET_INTERFACENAME,
						this.getClass().getClassLoader(), new PotentialFaultyPacketProxyHandler(packetDefinition)));
				lastPacketSent = lastPacketSent + packetDefinition.pauseInterval;
			}
		}
	}

	@Test
	public void testNoProblem() throws Exception, InterruptedException {
		ReceiverWrapper testReceiver = new ReceiverWrapper();
		testReceiver.initReceiver(10.0f);

		TestStream stream1 = new TestStream("0 F 100 F 100 . 100 . 100 .");

		stream1.simulateStream(testReceiver);
	}

	@Test(timeout = 10000, expected = IOException.class)
	public void testFaulty1() throws Exception, InterruptedException {
		ReceiverWrapper testReceiver = new ReceiverWrapper();
		testReceiver.initReceiver(1.0f);

		TestStream stream1 = new TestStream("0 F 100 F 100 . 100 . 100 .");

		stream1.simulateStream(testReceiver);
	}

	@Test(timeout = 10000)
	public void testAccurracy1() throws Exception, InterruptedException {
		ReceiverWrapper testReceiver = new ReceiverWrapper();
		testReceiver.initReceiver(2.0f);

		TestStream stream1 = new TestStream("0 . 10 F 500 F 100 . 100 . 100 .");

		stream1.simulateStream(testReceiver);
	}

	@Test(timeout = 10000, expected = IOException.class)
	public void testAccurracy2() throws Exception {
		ReceiverWrapper testReceiver = new ReceiverWrapper();
		testReceiver.initReceiver(2.0f);

		TestStream stream1 = new TestStream("0 . 10 F 500 F 100 F 100 . 100 .");

		stream1.simulateStream(testReceiver);
	}

	@Test(timeout = 10000)
	public void testAccurracy3() throws Exception, InterruptedException {
		ReceiverWrapper testReceiver = new ReceiverWrapper();
		testReceiver.initReceiver(2.0f);

		TestStream stream1 = new TestStream("0 F 500 F 200 . 310 F 100 .");

		stream1.simulateStream(testReceiver); // Gerade noch korrekt
	}

	@Test(timeout = 10000, expected = IOException.class)
	public void testAccurracy4() throws Exception, InterruptedException {
		ReceiverWrapper testReceiver = new ReceiverWrapper();
		testReceiver.initReceiver(2.0f);

		// dritter Fehler um 20 msec zu früh
		TestStream stream1 = new TestStream("0 F 500 F 200 . 280 F 100 .");

		stream1.simulateStream(testReceiver);
	}

}
